package com.sun.jndi.ldap;

import javax.naming.ConfigurationException;
import java.util.Vector;
import javax.naming.directory.InvalidAttributeIdentifierException;
import javax.naming.directory.InvalidAttributeValueException;
import javax.naming.directory.BasicAttribute;
import javax.naming.NamingEnumeration;
import javax.naming.directory.BasicAttributes;
import javax.naming.directory.DirContext;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;

final class LdapSchemaParser
{
    private static final boolean debug = false;
    static final String OBJECTCLASSDESC_ATTR_ID = "objectClasses";
    static final String ATTRIBUTEDESC_ATTR_ID = "attributeTypes";
    static final String SYNTAXDESC_ATTR_ID = "ldapSyntaxes";
    static final String MATCHRULEDESC_ATTR_ID = "matchingRules";
    static final String OBJECTCLASS_DEFINITION_NAME = "ClassDefinition";
    private static final String[] CLASS_DEF_ATTRS;
    static final String ATTRIBUTE_DEFINITION_NAME = "AttributeDefinition";
    private static final String[] ATTR_DEF_ATTRS;
    static final String SYNTAX_DEFINITION_NAME = "SyntaxDefinition";
    private static final String[] SYNTAX_DEF_ATTRS;
    static final String MATCHRULE_DEFINITION_NAME = "MatchingRule";
    private static final String[] MATCHRULE_DEF_ATTRS;
    private static final char SINGLE_QUOTE = '\'';
    private static final char WHSP = ' ';
    private static final char OID_LIST_BEGIN = '(';
    private static final char OID_LIST_END = ')';
    private static final char OID_SEPARATOR = '$';
    private static final String NUMERICOID_ID = "NUMERICOID";
    private static final String NAME_ID = "NAME";
    private static final String DESC_ID = "DESC";
    private static final String OBSOLETE_ID = "OBSOLETE";
    private static final String SUP_ID = "SUP";
    private static final String PRIVATE_ID = "X-";
    private static final String ABSTRACT_ID = "ABSTRACT";
    private static final String STRUCTURAL_ID = "STRUCTURAL";
    private static final String AUXILARY_ID = "AUXILIARY";
    private static final String MUST_ID = "MUST";
    private static final String MAY_ID = "MAY";
    private static final String EQUALITY_ID = "EQUALITY";
    private static final String ORDERING_ID = "ORDERING";
    private static final String SUBSTR_ID = "SUBSTR";
    private static final String SYNTAX_ID = "SYNTAX";
    private static final String SINGLE_VAL_ID = "SINGLE-VALUE";
    private static final String COLLECTIVE_ID = "COLLECTIVE";
    private static final String NO_USER_MOD_ID = "NO-USER-MODIFICATION";
    private static final String USAGE_ID = "USAGE";
    private static final String SCHEMA_TRUE_VALUE = "true";
    private boolean netscapeBug;
    
    LdapSchemaParser(boolean netscapeBug) {
        this.netscapeBug = netscapeBug;
    }
    
    static final void LDAP2JNDISchema(Attributes attributes, LdapSchemaCtx ldapSchemaCtx) throws NamingException {
        Attribute value = attributes.get("objectClasses");
        if (value != null) {
            objectDescs2ClassDefs(value, ldapSchemaCtx);
        }
        Attribute value2 = attributes.get("attributeTypes");
        if (value2 != null) {
            attrDescs2AttrDefs(value2, ldapSchemaCtx);
        }
        Attribute value3 = attributes.get("ldapSyntaxes");
        if (value3 != null) {
            syntaxDescs2SyntaxDefs(value3, ldapSchemaCtx);
        }
        Attribute value4 = attributes.get("matchingRules");
        if (value4 != null) {
            matchRuleDescs2MatchRuleDefs(value4, ldapSchemaCtx);
        }
    }
    
    private static final DirContext objectDescs2ClassDefs(Attribute attribute, LdapSchemaCtx ldapSchemaCtx) throws NamingException {
        BasicAttributes basicAttributes = new BasicAttributes(true);
        basicAttributes.put(LdapSchemaParser.CLASS_DEF_ATTRS[0], LdapSchemaParser.CLASS_DEF_ATTRS[1]);
        LdapSchemaCtx setup = ldapSchemaCtx.setup(2, "ClassDefinition", basicAttributes);
        NamingEnumeration<Object> all = attribute.getAll();
        while (all.hasMore()) {
            String s = (String)all.next();
            try {
                Object[] desc2Def = desc2Def(s);
                setup.setup(6, (String)desc2Def[0], (Attributes)desc2Def[1]);
            }
            catch (NamingException ex) {}
        }
        return setup;
    }
    
    private static final DirContext attrDescs2AttrDefs(Attribute attribute, LdapSchemaCtx ldapSchemaCtx) throws NamingException {
        BasicAttributes basicAttributes = new BasicAttributes(true);
        basicAttributes.put(LdapSchemaParser.ATTR_DEF_ATTRS[0], LdapSchemaParser.ATTR_DEF_ATTRS[1]);
        LdapSchemaCtx setup = ldapSchemaCtx.setup(3, "AttributeDefinition", basicAttributes);
        NamingEnumeration<Object> all = attribute.getAll();
        while (all.hasMore()) {
            String s = (String)all.next();
            try {
                Object[] desc2Def = desc2Def(s);
                setup.setup(7, (String)desc2Def[0], (Attributes)desc2Def[1]);
            }
            catch (NamingException ex) {}
        }
        return setup;
    }
    
    private static final DirContext syntaxDescs2SyntaxDefs(Attribute attribute, LdapSchemaCtx ldapSchemaCtx) throws NamingException {
        BasicAttributes basicAttributes = new BasicAttributes(true);
        basicAttributes.put(LdapSchemaParser.SYNTAX_DEF_ATTRS[0], LdapSchemaParser.SYNTAX_DEF_ATTRS[1]);
        LdapSchemaCtx setup = ldapSchemaCtx.setup(4, "SyntaxDefinition", basicAttributes);
        NamingEnumeration<Object> all = attribute.getAll();
        while (all.hasMore()) {
            String s = (String)all.next();
            try {
                Object[] desc2Def = desc2Def(s);
                setup.setup(8, (String)desc2Def[0], (Attributes)desc2Def[1]);
            }
            catch (NamingException ex) {}
        }
        return setup;
    }
    
    private static final DirContext matchRuleDescs2MatchRuleDefs(Attribute attribute, LdapSchemaCtx ldapSchemaCtx) throws NamingException {
        BasicAttributes basicAttributes = new BasicAttributes(true);
        basicAttributes.put(LdapSchemaParser.MATCHRULE_DEF_ATTRS[0], LdapSchemaParser.MATCHRULE_DEF_ATTRS[1]);
        LdapSchemaCtx setup = ldapSchemaCtx.setup(5, "MatchingRule", basicAttributes);
        NamingEnumeration<Object> all = attribute.getAll();
        while (all.hasMore()) {
            String s = (String)all.next();
            try {
                Object[] desc2Def = desc2Def(s);
                setup.setup(9, (String)desc2Def[0], (Attributes)desc2Def[1]);
            }
            catch (NamingException ex) {}
        }
        return setup;
    }
    
    private static final Object[] desc2Def(String s) throws NamingException {
        BasicAttributes basicAttributes = new BasicAttributes(true);
        int[] array = { 1 };
        int i = 1;
        Attribute numericOID = readNumericOID(s, array);
        String s2 = (String)numericOID.get(0);
        basicAttributes.put(numericOID);
        skipWhitespace(s, array);
        while (i != 0) {
            Attribute nextTag = readNextTag(s, array);
            basicAttributes.put(nextTag);
            if (nextTag.getID().equals("NAME")) {
                s2 = (String)nextTag.get(0);
            }
            skipWhitespace(s, array);
            if (array[0] < s.length() - 1) {
                continue;
            }
            i = 0;
        }
        return new Object[] { s2, basicAttributes };
    }
    
    private static final int findTrailingWhitespace(String s, int i) {
        while (i > 0) {
            if (s.charAt(i) != ' ') {
                return i + 1;
            }
            --i;
        }
        return 0;
    }
    
    private static final void skipWhitespace(String s, int[] array) {
        int i = array[0];
        while (i < s.length()) {
            if (s.charAt(i) != ' ') {
                array[0] = i;
                return;
            }
            ++i;
        }
    }
    
    private static final Attribute readNumericOID(String str, int[] array) throws NamingException {
        skipWhitespace(str, array);
        int n = array[0];
        int index = str.indexOf(32, n);
        if (index == -1 || index - n < 1) {
            throw new InvalidAttributeValueException("no numericoid found: " + str);
        }
        String substring = str.substring(n, index);
        int n2 = 0;
        array[n2] += substring.length();
        return new BasicAttribute("NUMERICOID", substring);
    }
    
    private static final Attribute readNextTag(String s, int[] array) throws NamingException {
        skipWhitespace(s, array);
        int index = s.indexOf(32, array[0]);
        String s2;
        if (index < 0) {
            s2 = s.substring(array[0], s.length() - 1);
        }
        else {
            s2 = s.substring(array[0], index);
        }
        String[] tag = readTag(s2, s, array);
        if (tag.length < 0) {
            throw new InvalidAttributeValueException("no values for attribute \"" + s2 + "\"");
        }
        BasicAttribute o = new BasicAttribute(s2, tag[0]);
        for (int i = 1; i < tag.length; ++i) {
            o.add(tag[i]);
        }
        return o;
    }
    
    private static final String[] readTag(String s, String s2, int[] array) throws NamingException {
        int n = 0;
        array[n] += s.length();
        skipWhitespace(s2, array);
        if (s.equals("NAME")) {
            return readQDescrs(s2, array);
        }
        if (s.equals("DESC")) {
            return readQDString(s2, array);
        }
        if (s.equals("EQUALITY") || s.equals("ORDERING") || s.equals("SUBSTR") || s.equals("SYNTAX")) {
            return readWOID(s2, array);
        }
        if (s.equals("OBSOLETE") || s.equals("ABSTRACT") || s.equals("STRUCTURAL") || s.equals("AUXILIARY") || s.equals("SINGLE-VALUE") || s.equals("COLLECTIVE") || s.equals("NO-USER-MODIFICATION")) {
            return new String[] { "true" };
        }
        if (s.equals("SUP") || s.equals("MUST") || s.equals("MAY") || s.equals("USAGE")) {
            return readOIDs(s2, array);
        }
        return readQDStrings(s2, array);
    }
    
    private static final String[] readQDString(String s, int[] array) throws NamingException {
        int n = s.indexOf(39, array[0]) + 1;
        int index = s.indexOf(39, n);
        if (n == -1 || index == -1 || n == index) {
            throw new InvalidAttributeIdentifierException("malformed QDString: " + s);
        }
        if (s.charAt(n - 1) != '\'') {
            throw new InvalidAttributeIdentifierException("qdstring has no end mark: " + s);
        }
        array[0] = index + 1;
        return new String[] { s.substring(n, index) };
    }
    
    private static final String[] readQDStrings(String s, int[] array) throws NamingException {
        return readQDescrs(s, array);
    }
    
    private static final String[] readQDescrs(String str, int[] array) throws NamingException {
        skipWhitespace(str, array);
        switch (str.charAt(array[0])) {
            case '(': {
                return readQDescrList(str, array);
            }
            case '\'': {
                return readQDString(str, array);
            }
            default: {
                throw new InvalidAttributeValueException("unexpected oids string: " + str);
            }
        }
    }
    
    private static final String[] readQDescrList(String str, int[] array) throws NamingException {
        Vector vector = new Vector(5);
        int n = 0;
        ++array[n];
        skipWhitespace(str, array);
        int i = array[0];
        int index = str.indexOf(41, i);
        if (index == -1) {
            throw new InvalidAttributeValueException("oidlist has no end mark: " + str);
        }
        while (i < index) {
            vector.addElement(readQDString(str, array)[0]);
            skipWhitespace(str, array);
            i = array[0];
        }
        array[0] = index + 1;
        String[] array2 = new String[vector.size()];
        for (int j = 0; j < array2.length; ++j) {
            array2[j] = (String)vector.elementAt(j);
        }
        return array2;
    }
    
    private static final String[] readWOID(String str, int[] array) throws NamingException {
        skipWhitespace(str, array);
        if (str.charAt(array[0]) == '\'') {
            return readQDString(str, array);
        }
        int n = array[0];
        int index = str.indexOf(32, n);
        if (index == -1 || n == index) {
            throw new InvalidAttributeIdentifierException("malformed OID: " + str);
        }
        array[0] = index + 1;
        return new String[] { str.substring(n, index) };
    }
    
    private static final String[] readOIDs(String str, int[] array) throws NamingException {
        skipWhitespace(str, array);
        if (str.charAt(array[0]) != '(') {
            return readWOID(str, array);
        }
        Vector vector = new Vector(5);
        int n = 0;
        ++array[n];
        skipWhitespace(str, array);
        int beginIndex = array[0];
        int index = str.indexOf(41, beginIndex);
        int n2 = str.indexOf(36, beginIndex);
        if (index == -1) {
            throw new InvalidAttributeValueException("oidlist has no end mark: " + str);
        }
        if (n2 == -1 || index < n2) {
            n2 = index;
        }
        while (n2 < index && n2 > 0) {
            vector.addElement(str.substring(beginIndex, findTrailingWhitespace(str, n2 - 1)));
            array[0] = n2 + 1;
            skipWhitespace(str, array);
            beginIndex = array[0];
            n2 = str.indexOf(36, beginIndex);
        }
        vector.addElement(str.substring(beginIndex, findTrailingWhitespace(str, index - 1)));
        array[0] = index + 1;
        String[] array2 = new String[vector.size()];
        for (int i = 0; i < array2.length; ++i) {
            array2[i] = (String)vector.elementAt(i);
        }
        return array2;
    }
    
    private final String classDef2ObjectDesc(Attributes attributes) throws NamingException {
        StringBuffer sb = new StringBuffer("( ");
        int n = 0;
        Attribute value = attributes.get("NUMERICOID");
        if (value != null) {
            sb.append(this.writeNumericOID(value));
            ++n;
            Attribute value2 = attributes.get("NAME");
            if (value2 != null) {
                sb.append(this.writeQDescrs(value2));
                ++n;
            }
            Attribute value3 = attributes.get("DESC");
            if (value3 != null) {
                sb.append(this.writeQDString(value3));
                ++n;
            }
            Attribute value4 = attributes.get("OBSOLETE");
            if (value4 != null) {
                sb.append(this.writeBoolean(value4));
                ++n;
            }
            Attribute value5 = attributes.get("SUP");
            if (value5 != null) {
                sb.append(this.writeOIDs(value5));
                ++n;
            }
            Attribute value6 = attributes.get("ABSTRACT");
            if (value6 != null) {
                sb.append(this.writeBoolean(value6));
                ++n;
            }
            Attribute value7 = attributes.get("STRUCTURAL");
            if (value7 != null) {
                sb.append(this.writeBoolean(value7));
                ++n;
            }
            Attribute value8 = attributes.get("AUXILIARY");
            if (value8 != null) {
                sb.append(this.writeBoolean(value8));
                ++n;
            }
            Attribute value9 = attributes.get("MUST");
            if (value9 != null) {
                sb.append(this.writeOIDs(value9));
                ++n;
            }
            Attribute value10 = attributes.get("MAY");
            if (value10 != null) {
                sb.append(this.writeOIDs(value10));
                ++n;
            }
            if (n < attributes.size()) {
                NamingEnumeration<Object> all = attributes.getAll();
                while (all.hasMoreElements()) {
                    Attribute attribute = (Attribute)all.next();
                    String id = attribute.getID();
                    if (id.equals("NUMERICOID") || id.equals("NAME") || id.equals("SUP") || id.equals("MAY") || id.equals("MUST") || id.equals("STRUCTURAL") || id.equals("DESC") || id.equals("AUXILIARY") || id.equals("ABSTRACT")) {
                        continue;
                    }
                    if (id.equals("OBSOLETE")) {
                        continue;
                    }
                    sb.append(this.writeQDStrings(attribute));
                }
            }
            sb.append(")");
            return sb.toString();
        }
        else {
            throw new ConfigurationException("Class definition doesn'thave a numeric OID");
        }
    }
    
    private final String attrDef2AttrDesc(Attributes attributes) throws NamingException {
        StringBuffer sb = new StringBuffer("( ");
        int n = 0;
        Attribute value = attributes.get("NUMERICOID");
        if (value != null) {
            sb.append(this.writeNumericOID(value));
            ++n;
            Attribute value2 = attributes.get("NAME");
            if (value2 != null) {
                sb.append(this.writeQDescrs(value2));
                ++n;
            }
            Attribute value3 = attributes.get("DESC");
            if (value3 != null) {
                sb.append(this.writeQDString(value3));
                ++n;
            }
            Attribute value4 = attributes.get("OBSOLETE");
            if (value4 != null) {
                sb.append(this.writeBoolean(value4));
                ++n;
            }
            Attribute value5 = attributes.get("SUP");
            if (value5 != null) {
                sb.append(this.writeWOID(value5));
                ++n;
            }
            Attribute value6 = attributes.get("EQUALITY");
            if (value6 != null) {
                sb.append(this.writeWOID(value6));
                ++n;
            }
            Attribute value7 = attributes.get("ORDERING");
            if (value7 != null) {
                sb.append(this.writeWOID(value7));
                ++n;
            }
            Attribute value8 = attributes.get("SUBSTR");
            if (value8 != null) {
                sb.append(this.writeWOID(value8));
                ++n;
            }
            Attribute value9 = attributes.get("SYNTAX");
            if (value9 != null) {
                sb.append(this.writeWOID(value9));
                ++n;
            }
            Attribute value10 = attributes.get("SINGLE-VALUE");
            if (value10 != null) {
                sb.append(this.writeBoolean(value10));
                ++n;
            }
            Attribute value11 = attributes.get("COLLECTIVE");
            if (value11 != null) {
                sb.append(this.writeBoolean(value11));
                ++n;
            }
            Attribute value12 = attributes.get("NO-USER-MODIFICATION");
            if (value12 != null) {
                sb.append(this.writeBoolean(value12));
                ++n;
            }
            Attribute value13 = attributes.get("USAGE");
            if (value13 != null) {
                sb.append(this.writeQDString(value13));
                ++n;
            }
            if (n < attributes.size()) {
                NamingEnumeration<Object> all = attributes.getAll();
                while (all.hasMoreElements()) {
                    Attribute attribute = (Attribute)all.next();
                    String id = attribute.getID();
                    if (id.equals("NUMERICOID") || id.equals("NAME") || id.equals("SYNTAX") || id.equals("DESC") || id.equals("SINGLE-VALUE") || id.equals("EQUALITY") || id.equals("ORDERING") || id.equals("SUBSTR") || id.equals("NO-USER-MODIFICATION") || id.equals("USAGE") || id.equals("SUP") || id.equals("COLLECTIVE")) {
                        continue;
                    }
                    if (id.equals("OBSOLETE")) {
                        continue;
                    }
                    sb.append(this.writeQDStrings(attribute));
                }
            }
            sb.append(")");
            return sb.toString();
        }
        else {
            throw new ConfigurationException("Attribute type doesn'thave a numeric OID");
        }
    }
    
    private final String syntaxDef2SyntaxDesc(Attributes attributes) throws NamingException {
        StringBuffer sb = new StringBuffer("( ");
        int n = 0;
        Attribute value = attributes.get("NUMERICOID");
        if (value != null) {
            sb.append(this.writeNumericOID(value));
            ++n;
            Attribute value2 = attributes.get("DESC");
            if (value2 != null) {
                sb.append(this.writeQDString(value2));
                ++n;
            }
            if (n < attributes.size()) {
                NamingEnumeration<Object> all = attributes.getAll();
                while (all.hasMoreElements()) {
                    Attribute attribute = (Attribute)all.next();
                    String id = attribute.getID();
                    if (id.equals("NUMERICOID")) {
                        continue;
                    }
                    if (id.equals("DESC")) {
                        continue;
                    }
                    sb.append(this.writeQDStrings(attribute));
                }
            }
            sb.append(")");
            return sb.toString();
        }
        else {
            throw new ConfigurationException("Attribute type doesn'thave a numeric OID");
        }
    }
    
    private final String matchRuleDef2MatchRuleDesc(Attributes attributes) throws NamingException {
        StringBuffer sb = new StringBuffer("( ");
        int n = 0;
        Attribute value = attributes.get("NUMERICOID");
        if (value != null) {
            sb.append(this.writeNumericOID(value));
            ++n;
            Attribute value2 = attributes.get("NAME");
            if (value2 != null) {
                sb.append(this.writeQDescrs(value2));
                ++n;
            }
            Attribute value3 = attributes.get("DESC");
            if (value3 != null) {
                sb.append(this.writeQDString(value3));
                ++n;
            }
            Attribute value4 = attributes.get("OBSOLETE");
            if (value4 != null) {
                sb.append(this.writeBoolean(value4));
                ++n;
            }
            Attribute value5 = attributes.get("SYNTAX");
            if (value5 != null) {
                sb.append(this.writeWOID(value5));
                ++n;
                if (n < attributes.size()) {
                    NamingEnumeration<Object> all = attributes.getAll();
                    while (all.hasMoreElements()) {
                        Attribute attribute = (Attribute)all.next();
                        String id = attribute.getID();
                        if (id.equals("NUMERICOID") || id.equals("NAME") || id.equals("SYNTAX") || id.equals("DESC")) {
                            continue;
                        }
                        if (id.equals("OBSOLETE")) {
                            continue;
                        }
                        sb.append(this.writeQDStrings(attribute));
                    }
                }
                sb.append(")");
                return sb.toString();
            }
            else {
                throw new ConfigurationException("Attribute type doesn'thave a syntax OID");
            }
        }
        else {
            throw new ConfigurationException("Attribute type doesn'thave a numeric OID");
        }
    }
    
    private final String writeNumericOID(Attribute attribute) throws NamingException {
        if (attribute.size() != 1) {
            throw new InvalidAttributeValueException("A class definition must have exactly one numeric OID");
        }
        return (String)(String)attribute.get() + 32;
    }
    
    private final String writeWOID(Attribute attribute) throws NamingException {
        if (this.netscapeBug) {
            return this.writeQDString(attribute);
        }
        return attribute.getID() + 32 + attribute.get() + 32;
    }
    
    private final String writeQDString(Attribute attribute) throws NamingException {
        if (attribute.size() != 1) {
            throw new InvalidAttributeValueException(attribute.getID() + " must have exactly one value");
        }
        return attribute.getID() + 32 + 39 + attribute.get() + 39 + 32;
    }
    
    private final String writeQDStrings(Attribute attribute) throws NamingException {
        return this.writeQDescrs(attribute);
    }
    
    private final String writeQDescrs(Attribute attribute) throws NamingException {
        switch (attribute.size()) {
            case 0: {
                throw new InvalidAttributeValueException(attribute.getID() + "has no values");
            }
            case 1: {
                return this.writeQDString(attribute);
            }
            default: {
                StringBuffer sb = new StringBuffer(attribute.getID());
                sb.append(' ');
                sb.append('(');
                NamingEnumeration<Object> all = attribute.getAll();
                while (all.hasMore()) {
                    sb.append(' ');
                    sb.append('\'');
                    sb.append((String)all.next());
                    sb.append('\'');
                    sb.append(' ');
                }
                sb.append(')');
                sb.append(' ');
                return sb.toString();
            }
        }
    }
    
    private final String writeOIDs(Attribute attribute) throws NamingException {
        switch (attribute.size()) {
            case 0: {
                throw new InvalidAttributeValueException(attribute.getID() + "has no values");
            }
            case 1: {
                if (this.netscapeBug) {
                    break;
                }
                return this.writeWOID(attribute);
            }
        }
        StringBuffer sb = new StringBuffer(attribute.getID());
        sb.append(' ');
        sb.append('(');
        NamingEnumeration<Object> all = attribute.getAll();
        sb.append(' ');
        sb.append(all.next());
        while (all.hasMore()) {
            sb.append(' ');
            sb.append('$');
            sb.append(' ');
            sb.append((String)all.next());
        }
        sb.append(' ');
        sb.append(')');
        sb.append(' ');
        return sb.toString();
    }
    
    private final String writeBoolean(Attribute attribute) throws NamingException {
        return attribute.getID() + 32;
    }
    
    final Attribute stringifyObjDesc(Attributes attributes) throws NamingException {
        BasicAttribute basicAttribute = new BasicAttribute("objectClasses");
        basicAttribute.add(this.classDef2ObjectDesc(attributes));
        return basicAttribute;
    }
    
    final Attribute stringifyAttrDesc(Attributes attributes) throws NamingException {
        BasicAttribute basicAttribute = new BasicAttribute("attributeTypes");
        basicAttribute.add(this.attrDef2AttrDesc(attributes));
        return basicAttribute;
    }
    
    final Attribute stringifySyntaxDesc(Attributes attributes) throws NamingException {
        BasicAttribute basicAttribute = new BasicAttribute("ldapSyntaxes");
        basicAttribute.add(this.syntaxDef2SyntaxDesc(attributes));
        return basicAttribute;
    }
    
    final Attribute stringifyMatchRuleDesc(Attributes attributes) throws NamingException {
        BasicAttribute basicAttribute = new BasicAttribute("matchingRules");
        basicAttribute.add(this.matchRuleDef2MatchRuleDesc(attributes));
        return basicAttribute;
    }
    
    static {
        CLASS_DEF_ATTRS = new String[] { "objectclass", "ClassDefinition" };
        ATTR_DEF_ATTRS = new String[] { "objectclass", "AttributeDefinition" };
        SYNTAX_DEF_ATTRS = new String[] { "objectclass", "SyntaxDefinition" };
        MATCHRULE_DEF_ATTRS = new String[] { "objectclass", "MatchingRule" };
    }
}
